import sys
from functools import reduce
from importlib import import_module
from pathlib import Path

import importlib.metadata
import importlib.resources

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.schema import ModelSchema, remove_star_keys
from oarepo_model_builder.utils.hyphen_munch import HyphenMunch


def create_builder_from_entrypoints(**kwargs):
    output_classes = load_entry_points_list("oarepo_model_builder.outputs")
    builder_classes = load_entry_points_list("oarepo_model_builder.builders")
    preprocess_classes = load_entry_points_list("oarepo_model_builder.property_preprocessors")
    model_preprocessor_classes = load_entry_points_list("oarepo_model_builder.model_preprocessors")

    builder_types = [x.TYPE for x in builder_classes]
    output_builder_components = {
        builder_type: load_entry_points_list(f"oarepo_model_builder.builder_components.{builder_type}")
        for builder_type in builder_types
    }

    return ModelBuilder(
        output_builders=builder_classes,
        outputs=output_classes,
        property_preprocessors=preprocess_classes,
        model_preprocessors=model_preprocessor_classes,
        output_builder_components=output_builder_components,
        **kwargs,
    )


def load_entry_points_dict(name):
    return {ep.name: ep.load() for ep in importlib.metadata.entry_points().select(group=name)}


def load_entry_points_list(name):
    ret = [(ep.name, ep.load()) for ep in importlib.metadata.entry_points().select(group=name)]
    ret.sort()
    return [x[1] for x in ret]


def load_model_from_entrypoint(ep: importlib.metadata.EntryPoint):
    def load(schema):
        try:
            loaded_schema = ep.load()
        except:
            module = import_module(ep.module)
            split_attr = ep.attr.split('.')
            fn = f'{split_attr[-2]}.{split_attr[-1]}'
            if len(split_attr) > 2:
                fn = reduce(lambda x, y: Path(x) / Path(y), split_attr[:-2]) / fn
            content = importlib.resources.open_text(module, fn, encoding='utf-8').read()
            loaded_schema = schema._load(fn, content=content)

        remove_star_keys(loaded_schema)
        return loaded_schema

    return load


def load_included_models_from_entry_points():
    ret = {}
    for ep in importlib.metadata.entry_points().select(group="oarepo.models"):
        ret[ep.name] = load_model_from_entrypoint(ep)
    return ret


def load_model(
        model_filename,
        package=None,
        configs=(),
        black=True,
        isort=True,
        sets=(),
        model_content=None,
        extra_included=None,
):
    loaders = load_entry_points_dict("oarepo_model_builder.loaders")
    included_models = load_included_models_from_entry_points()
    if extra_included:
        included_models.update(extra_included)
    schema = ModelSchema(
        model_filename,
        content=model_content,
        loaders=loaders,
        included_models=included_models,
    )
    for config in configs:
        load_config(schema, config, loaders)
    for s in sets:
        k, v = s.split("=", 1)
        schema.schema[k] = v
    check_plugin_packages(schema)
    if package:
        schema.settings["package"] = package
    if "python" not in schema.settings:
        schema.settings.python = HyphenMunch()
    schema.settings.python.use_isort = isort
    schema.settings.python.use_black = black
    return schema


def load_config(schema, config, loaders):
    old_loaders = schema.loaders
    schema.loaders = loaders
    try:
        loaded_file = schema._load(config)
        schema.merge(loaded_file)
    finally:
        schema.loaders = old_loaders


def check_plugin_packages(schema):
    try:
        required_packages = schema.schema.plugins.packages
    except AttributeError:
        return
    import subprocess

    import pkg_resources

    known_packages = set(d.project_name for d in pkg_resources.working_set)
    unknown_packages = [rp for rp in required_packages if rp not in known_packages]
    if unknown_packages:
        if (
                input(
                    f'Required packages {", ".join(unknown_packages)} are missing. '
                    f"Should I install them for you via pip install? (y/n) "
                )
                == "y"
        ):
            if subprocess.call(["pip", "install", *unknown_packages]):
                sys.exit(1)
            print("Installed required packages, please run this command again")
        sys.exit(1)



def deepmerge(target, source, stack=None, listmerge="overwrite"):
    if stack is None:
        stack = []

    if isinstance(target, dict):
        if source is not None:
            if not isinstance(source, dict):
                raise AttributeError(
                    f"Incompatible source and target on path {stack}: source {source}, target {target}"
                )
            for k, v in source.items():
                if k not in target:
                    target[k] = source[k]
                else:
                    target[k] = deepmerge(target[k], source[k], stack + [k], listmerge=listmerge)
    elif isinstance(target, list):
        if source is not None:
            if not isinstance(source, list):
                raise AttributeError(
                    f"Incompatible source and target on path {stack}: source {source}, target {target}"
                )
            if listmerge == "overwrite":
                for idx in range(min(len(source), len(target))):
                    target[idx] = deepmerge(target[idx], source[idx], stack + [idx], listmerge=listmerge)
                for idx in range(len(target), len(source)):
                    target.append(source[idx])
            elif listmerge == "extend":
                target.extend(source)
            elif listmerge == "keep":
                if len(source) > len(target):
                    target.extend(source[len(target):])
            else:
                raise AttributeError('listmerge must be one of "overwrite", "extend" or "keep"')
    return target



from datetime import date
from pathlib import Path

from os.path import basename
from oarepo_model_builder.builders import OutputBuilder
from oarepo_model_builder.builders.utils import ensure_directory
from ..utils.verbose import log
from cookiecutter.main import cookiecutter
from invenio_cli.helpers.cookiecutter_wrapper import CookiecutterWrapper
from invenio_cli.helpers.cli_config import CLIConfig


class CookiecutterBuilder(OutputBuilder):
    TYPE = "cookiecutter"

    def _generate_sample_site(self, app_package_path):
        package = self.schema.settings.package.replace('_', '-')
        template = 'gh:oarepo/cookiecutter-oarepo-instance'
        checkout = 'dev1'
        flavour = 'oarepo'

        builder = CookiecutterWrapper(
            flavour,
            **{
                'checkout': checkout,
                'no_input': True
            })
        builder.template_name = builder.extract_template_name(template)
        config_file = builder.create_and_dump_config_file()

        log.enter(log.INFO, 'Running cookiecutter...')
        project_dir = cookiecutter(
            template,
            checkout='dev1',
            config_file=config_file,
            overwrite_if_exists=True,
            no_input=True,
            extra_context=dict(
                project_name=f"{package} Sample site",
                project_shortname=f"{package}-sample-site",
                project_site="oarepo.org",
                github_repo=f"oarepo/{package}-sample-site",
                description=f"Testing OARepo Site instance for {package} datamodel.",
                author_name="CESNET",
                author_email="info@oarepo.org",
                year=f"{date.today().year}",
                python_version="3.9",
                database="postgresql",
                elasticsearch="7",
                file_storage="S3",
                development_tools="yes",
                app_package=f"{package}-sample-app",
                app_package_path=f"../{app_package_path}",
                # datamodel_package=package,
                # datamodel_package_path="../"
            )
        )
        saved_replay = builder.get_replay()
        CLIConfig.write(project_dir, flavour, saved_replay)
        log.leave(
            log.INFO, "Generated sample oarepo site in %s", project_dir)
        ensure_directory(self.builder, f"{Path(project_dir)}/logs")

    def _generate_sample_app(self):
        package = self.schema.settings.package.replace('_', '-')
        template = 'gh:oarepo/cookiecutter-oarepo-app'
        checkout = 'dev1'
        log.enter(log.INFO, 'Running cookiecutter...')
        project_dir = cookiecutter(
            template,
            checkout=checkout,
            overwrite_if_exists=True,
            no_input=True,
            extra_context=dict(
                project_name=f"{package} Sample app",
                project_shortname=f"{package}-sample-app",
                project_site="oarepo.org",
                github_repo=f"oarepo/{package}-sample-app",
                description=f"Testing OARepo application module with {package} datamodel.",
                author_name="CESNET",
                author_email="info@oarepo.org",
                year=f"{date.today().year}",
                python_version="3.9",
                database="postgresql",
                elasticsearch="7",
                file_storage="S3",
                development_tools="yes",
            )
        )
        log.leave(
            log.INFO, "Generated sample application in %s", project_dir)

        return basename(project_dir)

    def finish(self):
        super().finish()

        if (self.schema.settings.get('no-cookiecutter')):
            return

        app_project_dir = self._generate_sample_app()
        self._generate_sample_site(app_project_dir)
